"""
Plugin support for rdf.

There are a number of plugin points for rdf: parser, serializer,
store, query processor, and query result. Plugins can be registered
either through setuptools entry_points or by calling
rdf.plugin.register directly.

If you have a package that uses a setuptools based setup.py you can add the
following to your setup::

    entry_points = {
        'rdf.plugins.parser': [
            'nt =     rdf.plugins.parsers.nt:NTParser',
            ],
        'rdf.plugins.serializer': [
            'nt =     rdf.plugins.serializers.NTSerializer:NTSerializer',
            ],
        }

See the `setuptools dynamic discovery of services and plugins`__ for more
information.

.. __: http://peak.telecommunity.com/DevCenter/setuptools#dynamic-discovery-of-services-and-plugins

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from rdflib.store import Store
from rdflib.parser import Parser
from rdflib.serializer import Serializer
from rdflib.query import ResultParser, ResultSerializer, \
    Processor, Result, UpdateProcessor
from rdflib.exceptions import Error

__all__ = [
    'register', 'get', 'plugins', 'PluginException', 'Plugin', 'PKGPlugin']

entry_points = {'rdf.plugins.store': Store,
                'rdf.plugins.serializer': Serializer,
                'rdf.plugins.parser': Parser,
                'rdf.plugins.resultparser': ResultParser,
                'rdf.plugins.resultserializer': ResultSerializer,
                'rdf.plugins.queryprocessor': Processor,
                'rdf.plugins.queryresult': Result,
                'rdf.plugins.updateprocessor': UpdateProcessor
                }

_plugins = {}


class PluginException(Error):
    pass


class Plugin(object):

    def __init__(self, name, kind, module_path, class_name):
        self.name = name
        self.kind = kind
        self.module_path = module_path
        self.class_name = class_name
        self._class = None

    def getClass(self):
        if self._class is None:
            module = __import__(self.module_path, globals(), locals(), [""])
            self._class = getattr(module, self.class_name)
        return self._class


class PKGPlugin(Plugin):

    def __init__(self, name, kind, ep):
        self.name = name
        self.kind = kind
        self.ep = ep
        self._class = None

    def getClass(self):
        if self._class is None:
            self._class = self.ep.load()
        return self._class


def register(name, kind, module_path, class_name):
    """
    Register the plugin for (name, kind). The module_path and
    class_name should be the path to a plugin class.
    """
    p = Plugin(name, kind, module_path, class_name)
    _plugins[(name, kind)] = p


def get(name, kind):
    """
    Return the class for the specified (name, kind). Raises a
    PluginException if unable to do so.
    """
    try:
        p = _plugins[(name, kind)]
    except KeyError:
        raise PluginException(
            "No plugin registered for (%s, %s)" % (name, kind))
    return p.getClass()


try:
    from pkg_resources import iter_entry_points
except ImportError:
    pass  # TODO: log a message
else:
    # add the plugins specified via pkg_resources' EntryPoints.
    for entry_point, kind in entry_points.items():
        for ep in iter_entry_points(entry_point):
            _plugins[(ep.name, kind)] = PKGPlugin(ep.name, kind, ep)


def plugins(name=None, kind=None):
    """
    A generator of the plugins.

    Pass in name and kind to filter... else leave None to match all.
    """
    for p in _plugins.values():
        if (name is None or name == p.name) and (
                kind is None or kind == p.kind):
            yield p


register(
    'default', Store,
    'rdflib.plugins.memory', 'IOMemory')
register(
    'IOMemory', Store,
    'rdflib.plugins.memory', 'IOMemory')
register(
    'Auditable', Store,
    'rdflib.plugins.stores.auditable', 'AuditableStore')
register(
    'Concurrent', Store,
    'rdflib.plugins.stores.concurrent', 'ConcurrentStore')
register(
    'Sleepycat', Store,
    'rdflib.plugins.sleepycat', 'Sleepycat')
register(
    'SPARQLStore', Store,
    'rdflib.plugins.stores.sparqlstore', 'SPARQLStore')
register(
    'SPARQLUpdateStore', Store,
    'rdflib.plugins.stores.sparqlstore', 'SPARQLUpdateStore')

register(
    'application/rdf+xml', Serializer,
    'rdflib.plugins.serializers.rdfxml', 'XMLSerializer')
register(
    'xml', Serializer,
    'rdflib.plugins.serializers.rdfxml', 'XMLSerializer')
register(
    'text/n3', Serializer,
    'rdflib.plugins.serializers.n3', 'N3Serializer')
register(
    'n3', Serializer,
    'rdflib.plugins.serializers.n3', 'N3Serializer')
register(
    'text/turtle', Serializer,
    'rdflib.plugins.serializers.turtle', 'TurtleSerializer')
register(
    'turtle', Serializer,
    'rdflib.plugins.serializers.turtle', 'TurtleSerializer')
register(
    'ttl', Serializer,
    'rdflib.plugins.serializers.turtle', 'TurtleSerializer')
register(
    'trig', Serializer,
    'rdflib.plugins.serializers.trig', 'TrigSerializer')
register(
    'application/n-triples', Serializer,
    'rdflib.plugins.serializers.nt', 'NTSerializer')
register(
    'ntriples', Serializer,
    'rdflib.plugins.serializers.nt', 'NTSerializer')
register(
    'nt', Serializer,
    'rdflib.plugins.serializers.nt', 'NTSerializer')
register(
    'nt11', Serializer,
    'rdflib.plugins.serializers.nt', 'NT11Serializer')

register(
    'pretty-xml', Serializer,
    'rdflib.plugins.serializers.rdfxml', 'PrettyXMLSerializer')
register(
    'trix', Serializer,
    'rdflib.plugins.serializers.trix', 'TriXSerializer')
register(
    'application/trix', Serializer,
    'rdflib.plugins.serializers.trix', 'TriXSerializer')
register(
    'application/n-quads', Serializer,
    'rdflib.plugins.serializers.nquads', 'NQuadsSerializer')
register(
    'nquads', Serializer,
    'rdflib.plugins.serializers.nquads', 'NQuadsSerializer')

register(
    'application/rdf+xml', Parser,
    'rdflib.plugins.parsers.rdfxml', 'RDFXMLParser')
register(
    'xml', Parser,
    'rdflib.plugins.parsers.rdfxml', 'RDFXMLParser')
register(
    'text/n3', Parser,
    'rdflib.plugins.parsers.notation3', 'N3Parser')
register(
    'n3', Parser,
    'rdflib.plugins.parsers.notation3', 'N3Parser')
register(
    'text/turtle', Parser,
    'rdflib.plugins.parsers.notation3', 'TurtleParser')
register(
    'turtle', Parser,
    'rdflib.plugins.parsers.notation3', 'TurtleParser')
register(
    'ttl', Parser,
    'rdflib.plugins.parsers.notation3', 'TurtleParser')
register(
    'application/n-triples', Parser,
    'rdflib.plugins.parsers.nt', 'NTParser')
register(
    'ntriples', Parser,
    'rdflib.plugins.parsers.nt', 'NTParser')
register(
    'nt', Parser,
    'rdflib.plugins.parsers.nt', 'NTParser')
register(
    'nt11', Parser,
    'rdflib.plugins.parsers.nt', 'NTParser')
register(
    'application/n-quads', Parser,
    'rdflib.plugins.parsers.nquads', 'NQuadsParser')
register(
    'nquads', Parser,
    'rdflib.plugins.parsers.nquads', 'NQuadsParser')
register(
    'application/trix', Parser,
    'rdflib.plugins.parsers.trix', 'TriXParser')
register(
    'trix', Parser,
    'rdflib.plugins.parsers.trix', 'TriXParser')
register(
    'trig', Parser,
    'rdflib.plugins.parsers.trig', 'TrigParser')


register(
    'sparql', Result,
    'rdflib.plugins.sparql.processor', 'SPARQLResult')
register(
    'sparql', Processor,
    'rdflib.plugins.sparql.processor', 'SPARQLProcessor')

register(
    'sparql', UpdateProcessor,
    'rdflib.plugins.sparql.processor', 'SPARQLUpdateProcessor')


register(
    'xml', ResultSerializer,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultSerializer')
register(
    'application/sparql-results+xml', ResultSerializer,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultSerializer')
register(
    'txt', ResultSerializer,
    'rdflib.plugins.sparql.results.txtresults', 'TXTResultSerializer')
register(
    'json', ResultSerializer,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultSerializer')
register(
    'application/sparql-results+json', ResultSerializer,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultSerializer')
register(
    'csv', ResultSerializer,
    'rdflib.plugins.sparql.results.csvresults', 'CSVResultSerializer')
register(
    'text/csv', ResultSerializer,
    'rdflib.plugins.sparql.results.csvresults', 'CSVResultSerializer')

register(
    'xml', ResultParser,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultParser')
register(
    'application/sparql-results+xml', ResultParser,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultParser')
register(
    'application/sparql-results+xml; charset=UTF-8', ResultParser,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultParser')

register(
    'application/rdf+xml', ResultParser,
    'rdflib.plugins.sparql.results.graph', 'GraphResultParser')


register(
    'json', ResultParser,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultParser')
register(
    'application/sparql-results+json', ResultParser,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultParser')

register(
    'csv', ResultParser,
    'rdflib.plugins.sparql.results.csvresults', 'CSVResultParser')
register(
    'text/csv', ResultParser,
    'rdflib.plugins.sparql.results.csvresults', 'CSVResultParser')

register(
    'tsv', ResultParser,
    'rdflib.plugins.sparql.results.tsvresults', 'TSVResultParser')
register(
    'text/tab-separated-values', ResultParser,
    'rdflib.plugins.sparql.results.tsvresults', 'TSVResultParser')



from __future__ import absolute_import

from rdflib import ConjunctiveGraph
from rdflib.parser import Parser
from .notation3 import SinkParser, RDFSink


def becauseSubGraph(*args, **kwargs): pass


class TrigSinkParser(SinkParser):

    def directiveOrStatement(self, argstr, h):

        #import pdb; pdb.set_trace()

        i = self.skipSpace(argstr, h)
        if i < 0:
            return i    # EOF

        j = self.graph(argstr, i)
        if j >= 0:
            return j

        j = self.sparqlDirective(argstr, i)
        if j >= 0:
            return j

        j = self.directive(argstr, i)
        if j >= 0:
            return self.checkDot(argstr, j)

        j = self.statement(argstr, i)
        if j >= 0:
            return self.checkDot(argstr, j)

        return j

    def labelOrSubject(self, argstr, i, res):
        j = self.skipSpace(argstr, i)
        if j < 0:
            return j  # eof
        i = j

        j = self.uri_ref2(argstr, i, res)
        if j >= 0:
            return j

        if argstr[i] == '[':
            j = self.skipSpace(argstr, i + 1)
            if j < 0:
                self.BadSyntax(argstr, i,
                               "Expected ] got EOF")
            if argstr[j] == ']':
                res.append(self.blankNode())
                return j + 1
        return -1

    def graph(self, argstr, i):
        """
        Parse trig graph, i.e.

           <urn:graphname> = { .. triples .. }

        return -1 if it doesn't look like a graph-decl
        raise Exception if it looks like a graph, but isn't.
        """

        #import pdb; pdb.set_trace()
        j = self.sparqlTok('GRAPH', argstr, i)  # optional GRAPH keyword
        if j >= 0:
            i = j

        r = []
        j = self.labelOrSubject(argstr, i, r)
        if j >= 0:
            graph = r[0]
            i = j
        else:
            graph = self._store.graph.identifier  # hack

        j = self.skipSpace(argstr, i)
        if j < 0:
            self.BadSyntax(argstr, i,
                           "EOF found when expected graph")

        if argstr[j:j + 1] == "=":  # optional = for legacy support

            i = self.skipSpace(argstr, j + 1)
            if i < 0:
                self.BadSyntax(argstr, i, "EOF found when expecting '{'")
        else:
            i = j

        if argstr[i:i + 1] != "{":
            return -1  # the node wasn't part of a graph

        j = i + 1

        oldParentContext = self._parentContext
        self._parentContext = self._context
        reason2 = self._reason2
        self._reason2 = becauseSubGraph
        self._context = self._store.newGraph(graph)

        while 1:
            i = self.skipSpace(argstr, j)
            if i < 0:
                self.BadSyntax(
                    argstr, i, "needed '}', found end.")

            if argstr[i:i + 1] == "}":
                j = i + 1
                break

            j = self.directiveOrStatement(argstr, i)
            if j < 0:
                self.BadSyntax(
                    argstr, i, "expected statement or '}'")

        self._context = self._parentContext
        self._reason2 = reason2
        self._parentContext = oldParentContext
        # res.append(subj.close())    # No use until closed
        return j


class TrigParser(Parser):
    """
    An RDFLib parser for TriG

    """

    def __init__(self):
        pass

    def parse(self, source, graph, encoding="utf-8"):

        if encoding not in [None, "utf-8"]:
            raise Exception(
                ("TriG files are always utf-8 encoded, ",
                 "I was passed: %s") % encoding)

        # we're currently being handed a Graph, not a ConjunctiveGraph
        assert graph.store.context_aware, "TriG Parser needs a context-aware store!"

        conj_graph = ConjunctiveGraph(store=graph.store, identifier=graph.identifier)
        conj_graph.default_context = graph  # TODO: CG __init__ should have a
                                            # default_context arg
        # TODO: update N3Processor so that it can use conj_graph as the sink
        conj_graph.namespace_manager = graph.namespace_manager

        sink = RDFSink(conj_graph)

        baseURI = conj_graph.absolutize(
            source.getPublicId() or source.getSystemId() or "")
        p = TrigSinkParser(sink, baseURI=baseURI, turtle=True)

        p.loadStream(source.getByteStream())

        for prefix, namespace in p._bindings.items():
            conj_graph.bind(prefix, namespace)

        # return ???



"""
Trig RDF graph serializer for RDFLib.
See <http://www.w3.org/TR/trig/> for syntax specification.
"""

from collections import defaultdict

from rdflib.plugins.serializers.turtle import TurtleSerializer, _GEN_QNAME_FOR_DT, VERB
from rdflib.term import BNode, Literal
from six import b

__all__ = ['TrigSerializer']


class TrigSerializer(TurtleSerializer):

    short_name = "trig"
    indentString = 4 * u' '

    def __init__(self, store):
        if store.context_aware:
            self.contexts = list(store.contexts())
            self.default_context = store.default_context.identifier
            if store.default_context:
                self.contexts.append(store.default_context)
        else:
            self.contexts = [store]
            self.default_context = None

        super(TrigSerializer, self).__init__(store)

    def preprocess(self):
        for context in self.contexts:
            self.store = context
            self.getQName(context.identifier)
            self._references = defaultdict(int)
            self._subjects = {}

            for triple in context:
                self.preprocessTriple(triple)

            self._contexts[context] = (self.orderSubjects(), self._subjects, self._references)

    def reset(self):
        super(TrigSerializer, self).reset()
        self._contexts = {}

    def serialize(self, stream, base=None, encoding=None,
                  spacious=None, **args):
        self.reset()
        self.stream = stream
        # if base is given here, use that, if not and a base is set for the graph use that
        if base is not None:
            self.base = base
        elif self.store.base is not None:
            self.base = self.store.base

        if spacious is not None:
            self._spacious = spacious

        self.preprocess()

        self.startDocument()

        firstTime = True
        for store, (ordered_subjects, subjects, ref) in self._contexts.items():
            if not ordered_subjects:
                continue

            self._references = ref
            self._serialized = {}
            self.store = store
            self._subjects = subjects

            if self.default_context and store.identifier == self.default_context:
                self.write(self.indent() + '\n{')
            else:
                if isinstance(store.identifier, BNode):
                    iri = store.identifier.n3()
                else:
                    iri = self.getQName(store.identifier)
                    if iri is None:
                        iri = store.identifier.n3()
                self.write(self.indent() + '\n%s {' % iri)

            self.depth += 1
            for subject in ordered_subjects:
                if self.isDone(subject):
                    continue
                if firstTime:
                    firstTime = False
                if self.statement(subject) and not firstTime:
                    self.write('\n')
            self.depth -= 1
            self.write('}\n')

        self.endDocument()
        stream.write(b("\n"))



"""
This is a rdflib plugin for parsing NQuad files into Conjunctive
graphs that can be used and queried. The store that backs the graph
*must* be able to handle contexts.

>>> from rdflib import ConjunctiveGraph, URIRef, Namespace
>>> g = ConjunctiveGraph()
>>> data = open("test/nquads.rdflib/example.nquads", "rb")
>>> g.parse(data, format="nquads") # doctest:+ELLIPSIS
<Graph identifier=... (<class 'rdflib.graph.Graph'>)>
>>> assert len(g.store) == 449
>>> # There should be 16 separate contexts
>>> assert len([x for x in g.store.contexts()]) == 16
>>> # is the name of entity E10009 "Arco Publications"?
>>> #   (in graph http://bibliographica.org/entity/E10009)
>>> # Looking for:
>>> # <http://bibliographica.org/entity/E10009>
>>> #   <http://xmlns.com/foaf/0.1/name>
>>> #   "Arco Publications"
>>> #   <http://bibliographica.org/entity/E10009>
>>> s = URIRef("http://bibliographica.org/entity/E10009")
>>> FOAF = Namespace("http://xmlns.com/foaf/0.1/")
>>> assert(g.value(s, FOAF.name).eq("Arco Publications"))
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from codecs import getreader

from six import b

from rdflib import ConjunctiveGraph

# Build up from the NTriples parser:
from rdflib.plugins.parsers.ntriples import NTriplesParser
from rdflib.plugins.parsers.ntriples import ParseError
from rdflib.plugins.parsers.ntriples import r_tail
from rdflib.plugins.parsers.ntriples import r_wspace
from rdflib.plugins.parsers.ntriples import r_wspaces

__all__ = ['NQuadsParser']


class NQuadsParser(NTriplesParser):

    def parse(self, inputsource, sink, **kwargs):
        """Parse f as an N-Triples file."""
        assert sink.store.context_aware, ("NQuadsParser must be given"
                                          " a context aware store.")
        self.sink = ConjunctiveGraph(store=sink.store, identifier=sink.identifier)

        source = inputsource.getByteStream()

        if not hasattr(source, 'read'):
            raise ParseError("Item to parse must be a file-like object.")

        source = getreader('utf-8')(source)

        self.file = source
        self.buffer = ''
        while True:
            self.line = __line = self.readline()
            if self.line is None:
                break
            try:
                self.parseline()
            except ParseError as msg:
                raise ParseError("Invalid line (%s):\n%r" % (msg, __line))

        return self.sink

    def parseline(self):
        self.eat(r_wspace)
        if (not self.line) or self.line.startswith(('#')):
            return  # The line is empty or a comment

        subject = self.subject()
        self.eat(r_wspace)

        predicate = self.predicate()
        self.eat(r_wspace)

        obj = self.object()
        self.eat(r_wspace)

        context = self.uriref() or self.nodeid() or self.sink.identifier
        self.eat(r_tail)

        if self.line:
            raise ParseError("Trailing garbage")
        # Must have a context aware store - add on a normal Graph
        # discards anything where the ctx != graph.identifier
        self.sink.get_context(context).add((subject, predicate, obj))
